# Copyright 2021-2025 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transpiling circuits to IQM devices involving computational resonators.

In the IQM Star architecture, computational resonators are connected to multiple qubits.
A MOVE gate can be used to move the state of a qubit to a connected, empty computational resonator,
and back *to the same qubit*. Additionally, two-component gates like CZ can be applied
between a qubit and a connected resonator, so that effectively the qubit can be made to interact
with other qubits connected to the resonator. However, the resonators cannot be measured,
and no single-qubit gates can be applied on them.

To enable third-party transpilers to work on the IQM Star architecture, we may abstract away the
resonators and replace the real dynamic quantum architecture with a *simplified architecture*.
Specifically, this happens by removing the resonators from the architecture, and for
each resonator ``r``, and for each pair of supported native qubit-resonator (QR) gates
``(G(q1, r), MOVE(q2, r))`` adding the *fictional* gate ``G(q1, q2)`` to the simplified architecture.
This works since the fictional gate can be implemented as the QR gate sequence

::

    G(q1, q2) = [MOVE(q2, r), G(q1, r), MOVE(q2, r)].

This sequence is called a *resolution* of ``G(q1, q2)``.
Currently  we assume all the QR gates (other than MOVE) are symmetric in the sense that

::

    [MOVE(q2, r), G(q1, r), MOVE(q2, r)] = G(q1, q2) = G(q2, q1) = [MOVE(q1, r), G(q2, r), MOVE(q1, r)]

holds. This has the effect of doubling the number of possible resolutions for ``G(q1, q2)`` since
you can reverse the roles of the qubits.

Before a circuit transpiled to a simplified architecture can be executed it must be further
transpiled to the real Star architecture using :func:`transpile_insert_moves`, which will introduce
the resonators, add MOVE gates as necessary to move the states, and convert the fictional two-qubit
gates into real native gates acting on qubit-resonator pairs.

Likewise :func:`transpile_remove_moves` can be used to perform the opposite transformation,
converting a circuit valid for the real Star architecture into an equivalent circuit for the
corresponding simplified architecture, e.g. so that the circuit can be retranspiled or optimized
using third-party tools that do not support the MOVE gate.

Given a :class:`DynamicQuantumArchitecture` for a Star architecture, the corresponding simplified
version can be obtained using :func:`simplify_architecture`.
"""

from __future__ import annotations

from collections.abc import Collection, Iterable, Sequence
from enum import Enum

from iqm.iqm_client import (
    Circuit,
    CircuitTranspilationError,
    CircuitValidationError,
    DynamicQuantumArchitecture,
    Instruction,
    IQMClient,
)
from iqm.iqm_client.models import GateImplementationInfo, GateInfo, Locus, _op_is_symmetric

Resolution = tuple[str, str, str]
"""A (gate qubit, move qubit, resonator) triple that represents a resolution of a fictional
 qubit-qubit gate."""


class ExistingMoveHandlingOptions(str, Enum):
    """Options for how :func:`transpile_insert_moves` should handle existing MOVE instructions
    in the circuit.
    """

    KEEP = "keep"
    """Strict mode. The circuit, including existing MOVE instructions in it, is validated first.
    Then, any fictional two-qubit gates in the circuit are implemented with qubit-resonator gates."""
    TRUST = "trust"
    """Lenient mode. Same as KEEP, but does not validate the circuit first.
    Will attempt to fix any apparent user errors in the circuit by adding extra MOVE gates.
    """
    REMOVE = "remove"
    """Removes existing MOVE instructions from the circuit using :func:`transpile_remove_moves`, and
    then does the same as TRUST. This may produce a more optimized end result."""


def _map_loci(
    instructions: Iterable[Instruction],
    qubit_mapping: dict[str, str],
    inverse: bool = False,
) -> tuple[Instruction, ...]:
    """Map the loci of the given instructions using the given qubit mapping, or its inverse.

    Args:
        instructions: Instructions whose loci are to be mapped.
        qubit_mapping: Mapping from one set of qubit names to another. Assumed to be injective.
        inverse: Invert ``qubit_mapping`` before using it.

    Returns:
        Copies of ``instructions`` with mapped loci.

    """
    if inverse:
        qubit_mapping = {phys: log for log, phys in qubit_mapping.items()}
    return tuple(
        inst.model_copy(update={"qubits": tuple(qubit_mapping[q] for q in inst.qubits)}) for inst in instructions
    )


class _ResonatorStateTracker:
    r"""Tracks the qubit states stored in computational resonators on a Star architecture QPU
    as they are moved with MOVE gates.

    Since the MOVE gate is not defined when acting on a :math:`|11\rangle` state,
    and involves an unknown phase, it is not equivalent to a SWAP.
    The state must always be moved back to its original qubit (this reverses the unknown phase),
    and no gates may be applied on the qubit while its state is in a resonator (to avoid populating
    the :math:`|11\rangle` state).

    Also contains information on available qubit-resonator gate loci.

    Args:
        qr_gates: Mapping from qubit-resonator gate name to mapping from qubit to resonators with
            which it has the gate available.

    """

    move_gate = "move"
    """Name of the MOVE gate in the architecture."""
    qr_gate_names = frozenset(("move", "cz"))
    """Names of all arity-2 gates that *can* (in principle) be applied on a (qubit, resonator) locus
    in the real Star architecture. They *may* also have (qubit, qubit) loci available if the architecture
    allows it.
    Other arity-2 gates in the real architecture are assumed to *require* a (qubit, qubit) locus."""

    def __init__(self, qr_gates: dict[str, dict[str, set[str]]]) -> None:
        # NOTE: For now the transpilation logic assumes all qr gates other than MOVE are symmetric
        # in their effect, though not in implementation.
        for op in self.qr_gate_names:
            if op != self.move_gate and not _op_is_symmetric(op):
                raise ValueError(f"QR gate {op} is not symmetric.")

        def invert_locus_mapping(mapping: dict[str, set[str]]) -> dict[str, set[str]]:
            """Invert the give mapping of resonators to a list of connected qubits, returning
            a mapping of qubits to connected resonators (or vice versa).
            """
            inverse: dict[str, set[str]] = {}
            for r, qubits in mapping.items():
                for q in qubits:
                    inverse.setdefault(q, set()).add(r)
            return inverse

        self.move_q2r = qr_gates.pop(self.move_gate, {})
        """Mapping from qubit to resonators it can be MOVEd into."""
        self.move_r2q = invert_locus_mapping(self.move_q2r)
        """Mapping from resonator to qubits whose state can be MOVEd into it."""
        self.qr_gates_q2r = qr_gates
        """Mapping from QR gate name to mapping from qubit to resonators with which it has the gate available."""

        self.res_state_owner = {r: r for r in self.resonators}
        """Maps resonator to the QPU component whose state it currently holds."""

    @classmethod
    def from_dynamic_architecture(cls, arch: DynamicQuantumArchitecture) -> _ResonatorStateTracker:
        """Constructor to make a _ResonatorStateTracker from a dynamic quantum architecture.

        Args:
            arch: Architecture that determines the available gate loci.

        Returns:
            Tracker for ``arch``.

        """
        resonators = set(arch.computational_resonators)
        qr_gates: dict[str, dict[str, set[str]]] = {}
        for gate_name in cls.qr_gate_names:
            if gate_info := arch.gates.get(gate_name):
                qr_loci: dict[str, set[str]] = {}
                for q, r in gate_info.loci:
                    if gate_name == cls.move_gate:
                        # MOVE must have a (q, r) locus
                        if q in resonators or r not in resonators:
                            raise ValueError(f"MOVE gate locus {q, r} is not of the form (qubit, resonator)")
                    elif q in resonators:
                        # Other QR gates
                        raise ValueError(f"Gate {gate_name} locus {q, r} is not of the form (qubit, *)")
                    qr_loci.setdefault(q, set()).add(r)
                qr_gates[gate_name] = qr_loci
        return cls(qr_gates)

    @classmethod
    def from_circuit(cls, circuit: Circuit) -> _ResonatorStateTracker:
        """Constructor to make the _ResonatorStateTracker from a circuit.

        Infers the resonator connectivity and gate loci from the given the circuit.

        Args:
            circuit: The circuit to track the qubit states on.

        Returns:
            Tracker for the architecture inferred from ``circuit``.

        """
        return cls.from_instructions(circuit.instructions)

    @classmethod
    def from_instructions(cls, instructions: Iterable[Instruction]) -> _ResonatorStateTracker:
        """Constructor to make the _ResonatorStateTracker from a sequence of instructions.

        Infers the resonator connectivity from the MOVE gates.

        Args:
            instructions: The instructions to track the qubit states on.

        Returns:
            Tracker for the architecture inferred from the given instructions.

        """
        qr_gates: dict[str, dict[str, set[str]]] = {}
        for i in instructions:
            if i.name in cls.qr_gate_names:
                q, r = i.qubits
                qr_gates.setdefault(i.name, {}).setdefault(q, set()).add(r)
        return cls(qr_gates)

    @property
    def resonators(self) -> Collection[str]:
        """Computational resonators that are being tracked."""
        return self.move_r2q.keys()

    @property
    def qubit_state_holder(self) -> dict[str, str]:
        """Qubits not found in the dict hold their own states."""
        # TODO reverses res_state_owner, maybe this is the one we need to track instead?
        resonators = self.resonators
        # qubits whose states are in resonators
        return {c: r for r, c in self.res_state_owner.items() if c not in resonators}

    def apply_move(self, qubit: str, resonator: str) -> None:
        """Record changes to qubit state location when a MOVE gate is applied.

        Args:
            qubit: The moved qubit.
            resonator: The moved resonator.

        Raises:
            CircuitTranspilationError: MOVE is not allowed, either because the resonator does not exist,
                the MOVE gate is not available between this qubit-resonator pair, or the resonator is
                currently holding the state of a different qubit.

        """
        if (
            resonator in self.resonators
            and qubit in self.move_r2q[resonator]
            and (owner := self.res_state_owner[resonator]) in [qubit, resonator]
        ):
            self.res_state_owner[resonator] = qubit if owner == resonator else resonator
        else:
            raise CircuitTranspilationError(f"MOVE locus {qubit, resonator} is not allowed.")

    def create_move_instructions(
        self,
        qubit: str,
        resonator: str,
    ) -> Iterable[Instruction]:
        """MOVE instruction(s) to move the state of the given qubit into the given resonator,
        or back to the qubit.

        If the resonator has the state of another qubit in it, or if qubit's state is in another resonator,
        restore them first using additional MOVE instructions.

        Applies the returned MOVE instructions on the tracker state.

        Args:
            qubit: The qubit
            resonator: The resonator

        Yields:
            1, 2 or 3 MOVE instructions.

        """
        # if the resonator has another qubit's state in it, restore it
        owner = self.res_state_owner[resonator]
        if owner not in [qubit, resonator]:
            locus = (owner, resonator)
            self.apply_move(*locus)
            yield Instruction(name=self.move_gate, qubits=locus, args={})

        # if the qubit does not hold its own state, restore it, unless it's in the resonator
        # find where the qubit state is (it can be in at most one resonator)
        res = [r for r, c in self.res_state_owner.items() if c == qubit]  # TODO not efficient
        if res and (holder := res[0]) != resonator:
            locus = (qubit, holder)
            self.apply_move(*locus)
            yield Instruction(name=self.move_gate, qubits=locus, args={})

        # move qubit state to resonator, or back to the qubit
        locus = (qubit, resonator)
        self.apply_move(*locus)
        yield Instruction(name=self.move_gate, qubits=locus, args={})

    def restore_as_move_instructions(
        self,
        resonators: Iterable[str] | None = None,
    ) -> list[Instruction]:
        """MOVE instructions that move the states held in the given resonators back to their qubits.

        Applies the returned MOVE instructions on the tracker state.

        Args:
            resonators: Resonators that (may) hold qubit states that should be moved back to the qubits.
                If ``None``, the states in all known resonators will be returned to the qubits.

        Returns:
            MOVE instructions needed to move the states out of ``resonators`` into
            the qubits they belong to.

        """
        if resonators is None:
            resonators = self.resonators

        instructions: list[Instruction] = []
        for r in resonators:
            q = self.res_state_owner[r]
            # if the state in r does not belong to r, restore it to its owner
            if q != r:
                locus = (q, r)
                instructions.append(Instruction(name=self.move_gate, qubits=locus, args={}))
                self.apply_move(*locus)
        return instructions

    def resonators_holding_qubits(self, qubits: Iterable[str]) -> list[str]:
        """Return the resonators that are holding the state of one of the given qubits.

        Args:
            qubits: The qubits to check.

        Returns:
            Resonators that hold the state of one of ``qubits``.

        """
        # a resonator can only hold the state of a connected qubit, or its own state
        # TODO needs to be made more efficient once we have lots of resonators
        return [r for r, q in self.res_state_owner.items() if q != r and q in qubits]

    def map_resonators_in_locus(self, locus: Iterable[str]) -> Locus:
        """Map any resonators in the given instruction locus into the QPU components whose state is
        currently stored in that resonator.

        If the resonator contains no qubit state at the moment, it is not changed.
        Non-resonator components in the locus are also unchanged.

        Args:
            locus: Instruction locus to map.

        Returns:
            The mapped locus.

        """
        return tuple(self.res_state_owner.get(q, q) for q in locus)

    def find_resolutions(self, inst: Instruction) -> list[Resolution]:
        """Find all the possible resolutions for the given fictional qubit-qubit gate.

        Given a fictional gate G acting on qubits (a, b), finds all resonators r for which the current DQA
        has either G(a, r) and MOVE(b, r), or G(b, r) and MOVE(a, r) available. See :mod:`~iqm.iqm_client.transpile`.

        If ``inst`` is not a known QR gate, returns an empty list.

        Args:
            inst: Circuit instruction applying the fictional qubit-qubit gate G.

        Returns:
            All (gate qubit, move qubit, resonator) triples that can be used to implement ``inst``.

        """
        if (gate_q2r := self.qr_gates_q2r.get(inst.name)) is None:
            return []

        def get_resonators(g: str, m: str) -> set[str]:
            """Resonators r for which we have G(g, r) and MOVE(m, r) available."""
            return gate_q2r.get(g, set()) & self.move_q2r.get(m, set())

        # G is assumed symmetric, hence we may reverse the locus order for more resolutions
        a, b = inst.qubits
        return [(a, b, r) for r in get_resonators(a, b)] + [(b, a, r) for r in get_resonators(b, a)]

    def find_best_resolution(self, inst: Instruction, lookahead: Iterable[Instruction]) -> Resolution | None:
        """Find the best resolution for the fictional qubit-qubit gate instruction ``inst``
        using the available native qubit-resonator gates.

        Given a resolution (g, m, r) for ``inst``, it can be implemented as G(g, r) with
        additional MOVE gates applied first to make sure the state of the qubit m is in r,
        and g is holding its own state.

        Does not change the internal state of the tracker.

        Args:
            inst: instruction to implement
            lookahead: upcoming instructions, to be taken into consideration

        Returns:
            Best resolution for implementing ``inst``, or None iff no resolution could be found.

        """
        resolutions = self.find_resolutions(inst)
        if not resolutions:
            return None

        # We could sequence n ops of lookahead using recursion and use sequence lenght as badness,
        # but that would scale exponentially in n. Instead we do some heuristics:
        # Look ahead until we find the instructions that target the locus qubits next, and include
        # their costs in the badness calculation.
        followers: dict[str, Instruction] = {}
        for follower in lookahead:
            if len(followers) == 2:
                break  # found a follower for both locus qubits
            for q in follower.qubits:
                if q in inst.qubits:
                    followers.setdefault(q, follower)

        def get_badness(res: Resolution, g_holder: str, m_holder: str, r_owner: str) -> int:
            """Badness of the given resolution for implementing a fictional qubit-qubit gate.

            ``g_holder``, ``m_holder`` and ``r_owner`` represent the current QPU state.
            """
            g, m, r = res
            badness = 0
            if g_holder != g:
                badness += 1  # need to move the state back to g from a resonator
            if m_holder == r:
                pass  # m qubit state already in r
            else:
                if r_owner != r:
                    badness += 1  # resonator has some other qubit's state in it, and it must be restored
                if m_holder == m:
                    badness += 1  # need to move the m state to r
                else:
                    badness += 2  # m state is in another resonator, need 2 moves to get it to r
            return badness

        def follower_badness(res: Resolution) -> int:
            """Badness for implementing the follower instruction(s), starting from
            the situation after implementing ``inst`` itself using ``res``.
            """
            g, m, r = res
            g_follower = followers.get(g)
            m_follower = followers.get(m)

            if g_follower == m_follower and g_follower is not None:
                # >=2q gate. 2q gate has the same or reversed locus (no matter since QR gates are assumed symmetric!)
                if g_follower.name == inst.name:
                    # same gate => same resolution works, free
                    return 0
                follower_resolutions = self.find_resolutions(g_follower)
                if follower_resolutions:
                    # different resolvable QR gate
                    if res in follower_resolutions:
                        # same resolution works, free
                        return 0
                    # same resolution not ok, find the cheapest one
                    return min(get_badness(f_res, g, r, m) for f_res in follower_resolutions)
                # not a resolvable QR gate, ignore
                return 0

            badness: int = 0
            if g_follower:
                follower_resolutions = self.find_resolutions(g_follower)
                if follower_resolutions:
                    # g_follower is a 2q gate (sharing g only) that we can resolve
                    if any(f_res[2] != r for f_res in follower_resolutions):
                        # follower can be implemented using a different resonator
                        pass
                    else:
                        # follower needs to use same resonator but different move qubit
                        badness += 2
                else:
                    # Not a resolvable QR gate. For 1q gates on g, qubit state is already there.
                    pass

            if m_follower:
                follower_resolutions = self.find_resolutions(m_follower)
                if follower_resolutions:
                    # m_follower is a 2q gate (sharing m only) that we can resolve
                    if any(f_res[1:] == (m, r) for f_res in follower_resolutions):
                        # follower can be implemented using m as the move qubit and r as the resonator
                        pass
                    elif any(f_res[2] != r for f_res in follower_resolutions):
                        # follower can be implemented using a different resonator, m state must be restored
                        badness += 1
                    else:
                        # follower needs to use same resonator but different move qubit
                        badness += 2
                elif len(m_follower.qubits) == 1:
                    # Not a resolvable QR gate. For 1q gates on m, state must be restored.
                    badness += 1

            return badness

        options = []
        for res in resolutions:
            g, m, r = res
            # badness is the number of extra native instructions we need to implement ``inst``
            # and its followers using this resolution.
            # implementing ``inst`` itself, starting from the current tracker state
            badness: int = get_badness(
                res,
                self.qubit_state_holder.get(g, g),
                self.qubit_state_holder.get(m, m),
                self.res_state_owner[r],
            )

            # implementing the gate's followers, starting from the situation after implementing ``inst`` itself
            badness += follower_badness(res)

            if badness == 0:
                # badness cannot be lower than 0, so this is already an optimal resolution
                return res
            options.append((res, badness))

        # return the best option
        return min(options, key=lambda x: x[1])[0]

    def get_sequence(self, resolution: Resolution, inst: Instruction) -> list[Instruction]:
        """Apply a fictional two-qubit gate using a sequence of native qubit-resonator gates.

        See :mod:`~iqm.iqm_client.transpile`.
        Given the resolution (g, m, r), the gate is implemented as G(g, r), with additional MOVE gates
        applied first to make sure the state of the qubit m is in r, and g is holding its own state.

        Modifies the tracker state to reflect the returned sequence.

        Args:
            resolution: Defines a native gate sequence for implementing ``inst``.
            inst: Fictional qubit-qubit gate as an instruction.

        Returns:
            Sequence of real qubit-resonator gates implementing ``inst``.

        """
        g, m, r = resolution
        seq: list[Instruction] = []
        # does m state need to be moved to the resonator?
        m_holder = self.qubit_state_holder.get(m, m)
        if m_holder != r:
            seq += self.create_move_instructions(m, r)
        # does g state need to be moved to g?
        g_holder = self.qubit_state_holder.get(g, g)
        if g_holder != g:
            seq += self.restore_as_move_instructions([g_holder])
        # apply G(g, r)
        seq.append(inst.model_copy(update={"qubits": (g, r)}))
        return seq

    def insert_moves(
        self,
        instructions: Sequence[Instruction],
        arch: DynamicQuantumArchitecture,
    ) -> list[Instruction]:
        """Convert a simplified architecture circuit into a equivalent Star architecture circuit with
        resonators and MOVE gates.

        Inserts MOVE gates into the circuit and changes the existing instruction loci as needed,
        while updating the state of the tracker object.

        Can also handle circuits that mix the simplified and real architectures.

        Args:
            instructions: The instructions in the circuit, using physical qubit names.
            arch: Real Star quantum architecture we transpile to.

        Raises:
            CircuitTranspilationError: Raised when the circuit contains invalid gates that cannot be
                transpiled using this method.

        Returns:
            Real Star architecture equivalent of ``circuit`` with MOVEs and resonators added.

        """
        # This method can handle real single- and two-qubit gates, real q-r gates including MOVE,
        # and fictional two-qubit gates which it decomposes into real q-r gates.
        new_instructions: list[Instruction] = []

        for idx, inst in enumerate(instructions):
            locus = inst.qubits
            try:
                IQMClient._validate_instruction(architecture=arch, instruction=inst)
                # inst can be applied as is on locus, but we may first need to use MOVEs to make
                # sure the locus qubits contain their states

                if inst.name == self.move_gate:
                    # apply the requested MOVE, closing interfering MOVE sandwiches first
                    new_instructions += self.create_move_instructions(*locus)
                    continue

                # are some of the locus qubits' states currently in a resonator?
                if res_match := self.resonators_holding_qubits(locus):
                    # Some locus qubits do not hold their states, which need to be restored before applying the gate.
                    # NOTE: as a consequence, a barrier closes a MOVE sandwich.
                    new_instructions += self.restore_as_move_instructions(res_match)
                new_instructions.append(inst)

            except CircuitValidationError as e:
                # inst can not be applied to this locus as is
                if inst.name not in self.qr_gates_q2r or any(c in self.resonators for c in locus):
                    raise CircuitTranspilationError(e) from e

                resolution = self.find_best_resolution(inst, instructions[idx + 1 :])
                if resolution is None:
                    raise CircuitTranspilationError(
                        f"Unable to find native gate sequence to enable fictional gate {inst.name} at {locus}."
                        " Try routing the circuit to the simplified architecture first."
                    ) from e

                # implement G using the sequence
                new_instructions += self.get_sequence(resolution, inst)

        return new_instructions


def simplify_architecture(
    arch: DynamicQuantumArchitecture,
    *,
    remove_resonators: bool = True,
) -> DynamicQuantumArchitecture:
    """Converts the given IQM Star quantum architecture into the equivalent simplified quantum architecture.

    See :mod:`iqm.iqm_client.transpile` for the details.

    Adds fictional gates, abstracts away their gate implementations.
    Returns ``arch`` itself if it does not contain computational resonators (in which case nothing will change).

    Args:
        arch: quantum architecture to convert
        remove_resonators: iff False, return the union of the simplified and real architectures

    Returns:
        equivalent quantum architecture with fictional gates

    """
    # NOTE: assumes all qubit-resonator gates have the locus order (q, r)
    if not arch.computational_resonators:
        return arch

    r_set = frozenset(arch.computational_resonators)
    q_set = frozenset(arch.qubits)

    moves: dict[str, set[str]] = {}  # maps resonator r to qubits q for which we have MOVE(q, r) available
    for q, r in arch.gates["move"].loci if "move" in arch.gates else []:
        if q not in q_set or r not in r_set:
            raise ValueError(f"MOVE locus {q, r} is not of the form (qubit, resonator)")
        moves.setdefault(r, set()).add(q)

    def simplify_gate(gate_name: str, gate_info: GateInfo) -> GateInfo:
        """Convert the loci of the given gate"""
        new_loci: dict[str, tuple[Locus, ...]] = {}  # mapping from implementation to its new loci
        # loci for fictional gates, a set because multiple resonators can produce the same fictional locus
        fictional_loci: set[Locus] = set()

        for impl_name, impl_info in gate_info.implementations.items():
            kept_impl_loci: list[Locus] = []  # these real loci we keep for this implementation
            for locus in impl_info.loci:
                if len(locus) == 2:
                    # two-component op
                    q1, r = locus
                    if q1 not in q_set:
                        raise ValueError(f"Unexpected '{gate_name}' locus: {locus}")

                    if r in r_set:
                        if not remove_resonators:
                            kept_impl_loci.append(locus)
                        # involves a resonator, for each G(q1, r), MOVE(q2, r) pair add G(q1, q2) to the simplified arch
                        for q2 in moves.get(r, []):
                            if q1 != q2:
                                fictional_loci.add((q1, q2))
                    else:
                        # does not involve a resonator, keep
                        kept_impl_loci.append(locus)
                # other arities: keep as long as it does not involve a resonator
                elif not set(locus) & r_set:
                    kept_impl_loci.append(locus)
            new_loci[impl_name] = tuple(kept_impl_loci)

        # implementation info is lost in the simplification
        if fictional_loci:
            new_loci["__fictional"] = tuple(fictional_loci)

        return GateInfo(
            implementations={impl_name: GateImplementationInfo(loci=loci) for impl_name, loci in new_loci.items()},
            default_implementation=gate_info.default_implementation,
            override_default_implementation={
                locus: impl_name
                for locus, impl_name in gate_info.override_default_implementation.items()
                if all(c not in r_set for c in locus)
            },
        )

    # create fictional gates, remove real gate loci that involve a resonator
    new_gates: dict[str, GateInfo] = {}
    for gate_name, gate_info in arch.gates.items():
        if gate_name == "move":
            # MOVE gates do not have fictional versions
            if not remove_resonators:
                # keep the gate_info as is
                new_gates[gate_name] = gate_info
            continue
        new_gates[gate_name] = simplify_gate(gate_name, gate_info)

    return DynamicQuantumArchitecture(
        calibration_set_id=arch.calibration_set_id,
        qubits=arch.qubits,
        computational_resonators=[] if remove_resonators else arch.computational_resonators,
        gates=new_gates,
    )


def transpile_insert_moves(
    circuit: Circuit,
    arch: DynamicQuantumArchitecture,
    *,
    existing_moves: ExistingMoveHandlingOptions = ExistingMoveHandlingOptions.KEEP,
    qubit_mapping: dict[str, str] | None = None,
    restore_states: bool = True,
) -> Circuit:
    """Convert a simplified architecture circuit into an equivalent Star architecture circuit with
    resonators and MOVE gates, if needed.

    In the typical use case ``circuit`` has been transpiled to a simplified architecture
    where the resonators have been abstracted away, and this function converts it into
    the corresponding Star architecture circuit.

    It can also handle the case where ``circuit`` already contains MOVE gates and resonators,
    which are treated according to ``existing_moves``, followed by the conversion
    of the two-qubit gates that are not supported by the Star architecture.

    The function does nothing if ``arch`` does not support MOVE gates.

    Args:
        circuit: The circuit to convert.
        arch: Real Star architecture of the target device.
        existing_moves: Specifies how to deal with existing MOVE instructions in ``circuit``, if any.
        qubit_mapping: Mapping of logical qubit names to physical qubit names.
            Can be set to ``None`` if ``circuit`` already uses physical qubit names.
        restore_states: Iff True, all qubit states held in resonators are returned to their qubits
            at the end of the circuit (i.e. all MOVE sandwiches are closed), even when there
            is no computational reason to do so.

    Returns:
        Equivalent Star architecture circuit with MOVEs and resonators added.

    """
    move_gate = _ResonatorStateTracker.move_gate

    # see if the circuit already contains some MOVEs/resonators
    circuit_has_moves = any(i for i in circuit.instructions if i.name == move_gate)
    # can we use MOVEs?
    if move_gate not in arch.gates:
        if circuit_has_moves:
            raise ValueError("Circuit contains MOVE instructions, but the architecture does not support them.")
        # nothing to do (do not validate the circuit)
        return circuit

    tracker = _ResonatorStateTracker.from_dynamic_architecture(arch)

    # add missing QPU components to the mapping (mapped to themselves)
    if qubit_mapping is None:
        qubit_mapping = {}
    for c in set(arch.components) - set(qubit_mapping.values()):
        qubit_mapping[c] = c

    if existing_moves == ExistingMoveHandlingOptions.KEEP:
        # convert to physical qubit names
        phys_instructions = _map_loci(circuit.instructions, qubit_mapping)
        try:
            IQMClient._validate_circuit_moves(
                arch,
                Circuit(name=circuit.name, instructions=phys_instructions, metadata=circuit.metadata),
            )

        except CircuitValidationError as e:
            raise CircuitTranspilationError(e) from e
    else:
        if circuit_has_moves and existing_moves == ExistingMoveHandlingOptions.REMOVE:
            # convert the circuit into a pure simplified architecture circuit
            circuit = transpile_remove_moves(circuit)

        # convert to physical qubit names
        phys_instructions = _map_loci(circuit.instructions, qubit_mapping)

    new_instructions = tracker.insert_moves(phys_instructions, arch)

    if restore_states:
        new_instructions += tracker.restore_as_move_instructions()

    # convert back to logical qubit names
    return Circuit(
        name=circuit.name,
        instructions=_map_loci(new_instructions, qubit_mapping, inverse=True),
        metadata=circuit.metadata,
    )


def transpile_remove_moves(circuit: Circuit) -> Circuit:
    """Convert a Star architecture circuit involving resonators and MOVE gates into an equivalent
    simplified achitecture circuit without them.

    The method assumes that in ``circuit`` a MOVE gate is always used to move a qubit state into a
    resonator before any other gates act on the resonator. If this is not the case, this function
    will not work as intended.

    Args:
        circuit: Star architecture circuit from which resonators and MOVE gates should be removed.

    Returns:
        Equivalent simplified architecture circuit without resonators and MOVEs.

    """
    tracker = _ResonatorStateTracker.from_circuit(circuit)
    new_instructions = []
    for inst in circuit.instructions:
        if inst.name == tracker.move_gate:
            # update the state tracking, drop the MOVE
            tracker.apply_move(*inst.qubits)
        else:
            # map the instruction locus
            new_qubits = tracker.map_resonators_in_locus(inst.qubits)
            new_instructions.append(
                Instruction(name=inst.name, implementation=inst.implementation, qubits=new_qubits, args=inst.args)
            )
    return Circuit(name=circuit.name, instructions=tuple(new_instructions), metadata=circuit.metadata)



# Copyright 2021-2023 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command line interface for managing user authentication when using IQM quantum computers."""

import contextlib
from datetime import datetime, timedelta
import json
import logging
import os
from pathlib import Path
import platform
import sys
from typing import Any

import click
from iqm.iqm_client import __version__
from iqm.iqm_client.cli.auth import (
    AUTH_REQUESTS_TIMEOUT,
    ClientAccountSetupError,
    ClientAuthenticationError,
    login_request,
    logout_request,
    refresh_request,
    slash_join,
    time_left_seconds,
)
from iqm.iqm_client.cli.models import ConfigFile, TokensFile
from iqm.iqm_client.cli.token_manager import check_token_manager, daemonize_token_manager, start_token_manager
from psutil import NoSuchProcess, Process
from pydantic import ValidationError
import requests
from requests.exceptions import ConnectionError, Timeout

HOME_PATH = str(Path.home())
DEFAULT_CONFIG_PATH = os.path.join(HOME_PATH, ".config", "iqm-client-cli", "config.json")
DEFAULT_TOKENS_PATH = os.path.join(HOME_PATH, ".cache", "iqm-client-cli", "tokens.json")
REALM_NAME = "cortex"
CLIENT_ID = "iqm_client"
USERNAME = ""
REFRESH_PERIOD = 3 * 60  # in seconds


class ClickLoggingHandler(logging.Handler):
    """Simple log handler using click's echo function."""

    def __init__(self):
        super().__init__(level=logging.NOTSET)
        self.formatter = logging.Formatter("%(message)s")

    def emit(self, record):
        click.echo(self.format(record))


logger = logging.getLogger("iqm_client.cli.cli")
logger.addHandler(ClickLoggingHandler())
logger.setLevel(logging.INFO)


def _set_log_level_by_verbosity(verbose: bool) -> int:
    """Sets logger log level to DEBUG if verbose is True, to INFO otherwise.

    Args:
        verbose: whether logging should be verbose (i.e. DEBUG level)

    Returns:
        int: logging level which was set

    """
    if verbose:
        logger.setLevel(logging.DEBUG)
        return logging.DEBUG
    logger.setLevel(logging.INFO)
    return logging.INFO


class ResolvedPath(click.Path):
    """A click parameter type for a resolved path.
    Normal ``click.Path(resolve_path=True)`` fails under Windows running python <= 3.9.
    See https://github.com/pallets/click/issues/2466
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def convert(self, value: Any, param: click.Parameter | None, ctx: click.Context | None) -> Any:
        abspath = Path(value).absolute()
        # fsdecode to ensure that the return value is a str.
        # (with click<8.0.3 Path.convert will return Path if passed a Path)
        return os.fsdecode(super().convert(abspath, param, ctx))


def _read_json(path: str) -> dict:
    """Read a JSON file.

    Args:
        path: path to the file to read
    Raises:
        click.FileError: if file is not a valid JSON file
    Returns:
        dict: data parsed from the file

    """
    try:
        with open(path, "r", encoding="utf-8") as file:
            data = json.load(file)
    except FileNotFoundError as error:
        raise click.FileError(path, "file not found") from error
    except json.decoder.JSONDecodeError as error:
        raise click.FileError(path, f"file is not a valid JSON file: {error}") from error
    return data


def _is_parameter_source_cmd_line(ctx: click.Context, param_name: str | None) -> bool:
    """Check if the given parameter is submitted via command line"""
    if isinstance(param_name, str):
        param_source = ctx.get_parameter_source(param_name)
        if isinstance(param_source, click.core.ParameterSource):
            if param_source.name == "COMMANDLINE":
                return True
    return False


def _validate_path(ctx: click.Context, param: click.core.Option, path: str) -> str:
    """Callback for CLI prompt. If needed, confirmation to overwrite is prompted.

    The validation logic is not applied if the parameter is supplied via the
    command line.

    Args:
        ctx: click context
        param: click prompt param object
        path: path provided by user
    Returns:
        str: confirmed and finalized path

    """
    if ctx.obj is None:
        ctx.obj = {}
    if param.name in ctx.obj:
        return path
    ctx.obj[param.name] = True

    # Skip parameter validation completely if the parameter comes from command line
    if _is_parameter_source_cmd_line(ctx, param.name):
        msg = click.style(
            f'Skipping validation of "{param.opts[0]}", using the provided value "{path}" as is',
            fg="yellow",
        )
        click.echo(msg)
        return path

    # File doesn't exist, no need to confirm overwriting
    if not Path(path).is_file():
        return path

    # File exists, so user must either overwrite or enter a new path
    while True:
        msg = f"{click.style('File at given path already exists. Overwrite?', fg='red')}"
        if click.confirm(msg, default=None):
            return path

        new_path = click.prompt("New file path", type=ResolvedPath(dir_okay=False, writable=True, resolve_path=True))

        if new_path == path:
            continue
        return new_path


def _validate_config_file(config_file: str) -> ConfigFile:
    """Checks if provided config file is valid, i.e. it:
       - is valid JSON
       - satisfies IQM Client CLI format

    Args:
        config_file (str): --config-file option value
    Raises:
        click.FileError: if config_file is not valid JSON
        click.FileError: if config_file does not satisfy IQM Client CLI format
    Returns:
        ConfigFile: validated config loaded from config_file

    """
    # config_file must be in correct format
    config = _read_json(config_file)
    try:
        validated_config = ConfigFile(**config)
    except ValidationError as ex:
        raise click.FileError(
            config_file,
            f"""Provided config file is valid JSON, but does not satisfy IQM Client CLI format. Possible reasons:
- IQM Client CLI was upgraded and config file format is changed. Check the changelog.
- Config file was manually edited by someone.

Re-generate a valid config file by running 'iqmclient init'.

Full validation error:
{ex}""",
        )

    return validated_config


def _validate_tokens_file(tokens_file: str) -> TokensFile:
    """Checks if provided tokens file is valid, i.e. it:
       - is valid JSON
       - satisfies IQM Client CLI format

    Args:
        tokens_file (str): path to tokens file
    Raises:
        click.FileError: if tokens file is not valid JSON
        click.FileError: if tokens file does not satisfy IQM Client CLI format
    Returns:
        TokensFile: validated tokens loaded from tokens_file

    """
    # tokens_file must be in correct format
    tokens = _read_json(tokens_file)
    try:
        validated_tokens = TokensFile(**tokens)
    except ValidationError as ex:
        raise click.FileError(
            tokens_file,
            f"""Provided tokens file is valid JSON, but does not satisfy IQM Client CLI format. Possible reasons:
- IQM Client CLI was upgraded and tokens file format is changed. Check the changelog.
- Tokens file was manually edited by someone.

Re-generate a valid tokens file by running 'iqmclient auth login'.

Full validation error:
{ex}""",
        )

    return validated_tokens


def _validate_auth_server_url(ctx: click.Context, param: click.Option, base_url: str) -> str:
    """Checks if provided auth server URL is valid, i.e. it:
       - is a valid HTTP/HTTPS URL
       - is accessible
       - points to an authentication server

    The validation logic is not applied if the parameter is supplied via the
    command line.

    Args:
        ctx: click context
        param: click prompt param object
        base_url (str): auth server base URL to validate
    Returns:
        str: validated auth server base URL

    """
    if ctx.obj is None:
        ctx.obj = {}
    if param.name in ctx.obj:
        return base_url

    if _is_parameter_source_cmd_line(ctx, param.name):
        msg = click.style(
            f'Skipping validation of "{param.opts[0]}", using the provided value "{base_url}" as is',
            fg="yellow",
        )
        click.echo(msg)
        ctx.obj[param.name] = base_url
        return base_url

    is_valid = False
    while not is_valid:
        try:
            master = requests.get(f"{base_url}/realms/master", timeout=AUTH_REQUESTS_TIMEOUT)
            assert master.status_code == 200
            assert "public_key" in master.json()
            is_valid = True
        except (ConnectionError, AssertionError, ValueError):
            click.echo(f"No auth server could be accessed with URL {base_url}")
            is_valid = click.confirm("Do you still want to use it?", default=False)
        if not is_valid:
            base_url = click.prompt(str(param.prompt))

    ctx.obj[param.name] = base_url
    return base_url


def _validate_auth_realm(ctx: click.Context, param: click.Option, realm: str) -> str:
    """Checks if provided realm exists on auth server.

    The validation logic is not applied if the parameter is supplied via the
    command line.

    Args:
        ctx: click context
        param: click prompt param object
        realm (str): name of the realm
    Returns:
        str: validated realm name

    """
    if ctx.obj is None:
        ctx.obj = {}
    if param.name in ctx.obj:
        return realm

    if _is_parameter_source_cmd_line(ctx, param.name):
        msg = click.style(
            f'Skipping validation of "{param.opts[0]}", using the provided value "{realm}" as is',
            fg="yellow",
        )
        click.echo(msg)
        ctx.obj[param.name] = realm
        return realm

    base_url = ctx.obj.get("auth_server_url", None)
    if base_url is None:
        raise click.UsageError("Can not set realm name before setting auth server URL.")

    is_valid = False
    while not is_valid:
        try:
            realm_data = requests.get(f"{base_url}/realms/{realm}", timeout=AUTH_REQUESTS_TIMEOUT)
            assert realm_data.status_code == 200
            assert "public_key" in realm_data.json()
            is_valid = True
        except (ConnectionError, AssertionError, ValueError):
            click.echo(f"No auth realm could be accessed with URL {base_url}/realms/{realm}")
            is_valid = click.confirm("Do you still want to use it?", default=False)
        if not is_valid:
            realm = click.prompt(str(param.prompt))

    ctx.obj[param.name] = realm
    return realm


class IQMClientCliCommand(click.Group):
    """A custom click command group class to wrap global constants."""

    default_config_path: str = DEFAULT_CONFIG_PATH
    default_tokens_path: str = DEFAULT_TOKENS_PATH


@click.group(cls=IQMClientCliCommand)
@click.version_option(__version__)
def iqmclient() -> None:
    """IQM Client CLI for managing user authentication when using IQM quantum computers"""
    return


@iqmclient.command()
@click.help_option()
@click.option(
    "--config-file",
    prompt="Where to save config",
    callback=_validate_path,
    default=IQMClientCliCommand.default_config_path,
    type=ResolvedPath(dir_okay=False, writable=True, resolve_path=True),
    is_eager=True,
    help="Location where the configuration file will be saved.",
)
@click.option(
    "--tokens-file",
    prompt="Where to save auth tokens",
    callback=_validate_path,
    default=IQMClientCliCommand.default_tokens_path,
    type=ResolvedPath(dir_okay=False, writable=True, resolve_path=True),
    is_eager=True,
    help="Location where the tokens file will be saved.",
)
@click.option(
    "--auth-server-url",
    prompt="Authentication server URL",
    callback=_validate_auth_server_url,
    is_eager=True,
    help="Authentication server URL.",
)
@click.option(
    "--realm",
    prompt="Realm on IQM auth server",
    prompt_required=False,
    default=REALM_NAME,
    show_default=True,
    callback=_validate_auth_realm,
    help="Name of the realm on the IQM authentication server.",
)
@click.option(
    "--client-id",
    prompt="Client ID",
    prompt_required=False,
    default=CLIENT_ID,
    show_default=True,
    help="Client ID on the IQM authentication server.",
)
@click.option(
    "--username",
    prompt="Username (optional)",
    required=False,
    default=USERNAME,
    help="Username. If not provided, it will be asked for at login.",
)
@click.option("-v", "--verbose", is_flag=True, help="Print extra information.")
def init(
    config_file: str, tokens_file: str, auth_server_url: str, realm: str, client_id: str, username: str, verbose: bool
) -> None:
    """Initialize configuration and authentication."""
    _set_log_level_by_verbosity(verbose)

    path_to_dir = Path(config_file).parent
    config_json = json.dumps(
        {
            "auth_server_url": auth_server_url,
            "realm": realm,
            "client_id": client_id,
            "username": username,
            "tokens_file": tokens_file,
        },
        indent=2,
    )

    # Tokens file exist, so token manager may be running. Notify user and kill token manager.
    if Path(tokens_file).is_file():
        pid = check_token_manager(tokens_file)
        if pid:
            logger.info("Active token manager (PID %s) will be killed.", pid)
            _safe_process_terminate(pid)

        # Remove tokens file to start from scratch after init
        os.remove(tokens_file)

    try:
        path_to_dir.mkdir(parents=True, exist_ok=True)
        with open(Path(config_file), "w", encoding="UTF-8") as file:
            file.write(config_json)
            logger.debug("Saved configuration file: %s", config_file)
    except OSError as error:
        raise click.ClickException(f"Error writing configuration file, {error}") from error

    logger.info(
        "IQM Client CLI initialized successfully. Login and start the token manager with 'iqmclient auth login'."
    )


@iqmclient.group()
def auth() -> None:
    """Manage authentication."""
    return


@auth.command()
@click.option(
    "--config-file",
    default=IQMClientCliCommand.default_config_path,
    type=ResolvedPath(exists=True, dir_okay=False, resolve_path=True),
    help="Location of the configuration file to be used.",
)
@click.option("-v", "--verbose", is_flag=True, help="Print extra information.")
def status(config_file, verbose) -> None:
    """Check status of authentication."""
    _set_log_level_by_verbosity(verbose)

    logger.debug("Using configuration file: %s", config_file)
    config = _validate_config_file(config_file)
    tokens_file = str(config.tokens_file)
    if not config.tokens_file.is_file():
        click.echo('Not logged in. Use "iqmclient auth login" to login.')
        return
    try:
        tokens_data = _validate_tokens_file(tokens_file)
    except click.FileError:
        click.echo('Provided tokens.json file is invalid. Use "iqmclient auth login" to generate new tokens.')
        return

    click.echo(f"Tokens file: {tokens_file}")
    if not tokens_data.pid:
        click.echo(
            "Tokens file doesn't contain PID. Probably, 'iqmclient auth login' was launched with '--no-refresh'\n"
        )

    refresh_status = tokens_data.refresh_status or "SUCCESS"
    styled_status = click.style(refresh_status, fg="green" if refresh_status == "SUCCESS" else "red")
    refresh_timestamp = tokens_data.timestamp.strftime("%m/%d/%Y %H:%M:%S")
    click.echo(f"Last refresh: {refresh_timestamp} from {tokens_data.auth_server_url} {styled_status}")
    seconds_at = time_left_seconds(tokens_data.access_token)
    time_left_at = str(timedelta(seconds=seconds_at))
    click.echo(f"Time left on access token (hh:mm:ss): {time_left_at}")
    seconds_rt = time_left_seconds(tokens_data.refresh_token)
    time_left_rt = str(timedelta(seconds=seconds_rt))
    click.echo(f"Time left on refresh token (hh:mm:ss): {time_left_rt}")

    active_pid = check_token_manager(tokens_file)
    if active_pid:
        click.echo(f"Token manager: {click.style('RUNNING', fg='green')} (PID {active_pid})")
    else:
        click.echo(f"Token manager: {click.style('NOT RUNNING', fg='red')}")


def _validate_iqm_client_cli_auth_login(no_daemon, no_refresh, config_file) -> ConfigFile:
    """Checks if provided combination of auth login options is valid:
       - no_daemon and no_refresh are mutually exclusive
       - config file should pass validation

    Args:
        config_file (str): --config-file option value
        no_daemon (bool): --no-daemon option value
        no_refresh (bool): --no-refresh option value
    Raises:
        click.BadOptionUsage: if both mutually exclusive --no-daemon and --no-refresh are set
        click.BadParameter: if config_file does not exist
    Returns:
        ConfigFile: validated config loaded from config_file

    """
    # --no-refresh and --no-daemon are mutually exclusive
    if no_refresh and no_daemon:
        raise click.BadOptionUsage(
            "--no-refresh", "Cannot request a non-daemonic (foreground) token manager when using '--no-refresh'."
        )

    # config file, even the default one, should exist
    if not Path(config_file).is_file():
        raise click.BadParameter(
            f"Provided config {config_file} does not exist. "
            + "Provide a different file or run 'iqmclient auth init' to create a new config file."
        )

    # config file should be valid JSON and satisfy IQM Client CLI format
    config = _validate_config_file(config_file)

    return config


def _refresh_tokens(
    refresh_period: int,
    no_daemon: bool,
    no_refresh: bool,
    config: ConfigFile,
) -> bool:
    """Refreshes token and returns success status

    Args:
        refresh_period (int): --refresh_period option value
        no_daemon (bool): --no-daemon option value
        no_refresh (bool): --no-refresh option value
        config (ConfigFile): IQM Client CLI config
    Returns:
        bool: whether token refresh was successful or not

    """
    # Tokens file exists; Refresh tokens without username/password
    tokens_file = str(config.tokens_file)
    try:
        refresh_token = _validate_tokens_file(tokens_file).refresh_token
    except (click.FileError, ValidationError):
        click.echo("Provided tokens.json file is invalid, continuing with login with username and password.")
        os.remove(tokens_file)
        return False

    logger.debug("Attempting to refresh tokens by using existing refresh token from file: %s", tokens_file)

    new_tokens = None
    try:
        new_tokens = refresh_request(str(config.auth_server_url), config.realm, config.client_id, refresh_token)
    except (Timeout, ConnectionError, ClientAuthenticationError):
        logger.info("Failed to refresh tokens by using existing token. Switching to username/password.")

    if new_tokens:
        save_tokens_file(tokens_file, new_tokens, str(config.auth_server_url))
        logger.debug("Saved new tokens file: %s", tokens_file)
        if no_refresh:
            logger.info("Existing token used to refresh session. Token manager not started due to '--no-refresh' flag.")
        elif no_daemon:
            logger.info("Existing token was used to refresh the auth session. Token manager started in foreground...")
            start_token_manager(refresh_period, config)
        else:
            logger.info("Existing token was used to refresh the auth session. Token manager daemon started.")
            daemonize_token_manager(refresh_period, config)
        return True
    return False


@auth.command()
@click.option(
    "--config-file",
    default=IQMClientCliCommand.default_config_path,
    type=ResolvedPath(exists=True, dir_okay=False, resolve_path=True),
    help="Location of the configuration file to be used.",
)
@click.option("--username", help="Username for authentication.")
@click.option("--password", help="Password for authentication.")
@click.option(
    "--refresh-period", default=REFRESH_PERIOD, show_default=True, help="How often to refresh tokens (in seconds)."
)
@click.option("--no-daemon", is_flag=True, default=False, help="Start token manager in foreground, not as daemon.")
@click.option(
    "--no-refresh", is_flag=True, default=False, help="Login, but do not start token manager to refresh tokens."
)
@click.option("-v", "--verbose", is_flag=True, help="Print extra information.")
def login(
    config_file: str,
    username: str,
    password: str,
    refresh_period: int,
    no_daemon: bool,
    no_refresh: bool,
    verbose: bool,
) -> None:
    """Authenticate on the IQM server, and optionally start a token manager to maintain the session."""
    _set_log_level_by_verbosity(verbose)

    if platform.system().lower().startswith("win") and not no_refresh and not no_daemon:
        click.echo(
            click.style("Warning", fg="yellow")
            + ": Daemonizing is not supported on Windows, and the application has started in foreground mode; "
            "please keep this terminal session open in order for IQM Client CLI to keep refreshing the tokens and "
            "maintaining the authentication.\n"
        )
        no_daemon = True

    # Validate whether the combination of options makes sense
    config = _validate_iqm_client_cli_auth_login(no_daemon, no_refresh, config_file)

    auth_server_url, realm, client_id = str(config.auth_server_url), config.realm, config.client_id
    tokens_file = str(config.tokens_file)

    if config.tokens_file.is_file():
        if check_token_manager(tokens_file):
            logger.info("Login aborted, because token manager is already running. See 'iqmclient auth status'.")
            return

        if _refresh_tokens(refresh_period, no_daemon, no_refresh, config):
            return

    # Login with username and password
    username = username or config.username or click.prompt("Username")
    if config.username:
        click.echo(f"Username: {username}")
    password = password or click.prompt("Password", hide_input=True)
    tokens = None

    while tokens is None:
        try:
            tokens = login_request(auth_server_url, realm, client_id, username, password)
        except ConnectionError as exc:
            raise click.ClickException(f"Authentication server at {auth_server_url} is not accessible") from exc
        except Timeout as exc:
            raise click.ClickException(f"Authentication server at {auth_server_url} is not responding") from exc
        except ClientAuthenticationError as exc:
            raise click.ClickException(f"Failed to authenticate, {exc}") from exc
        except ClientAccountSetupError as exc:
            password_update_form_url = slash_join(auth_server_url, f"realms/{realm}/account")
            raise click.ClickException(
                f"""
Failed to authenticate, because your account is not fully set up yet.
Please update your password at {password_update_form_url}
"""
            ) from exc

    logger.info("Logged in successfully as %s", username)
    save_tokens_file(tokens_file, tokens, auth_server_url)
    env_var_command = "set" if platform.system().lower().startswith("win") else "export"
    click.echo(
        f"""
To use the tokens file with IQM Client or IQM Client-based software, set the environment variable:

{env_var_command} IQM_TOKENS_FILE={tokens_file}

Refer to IQM Client documentation for details: https://docs.meetiqm.com/iqm-client/
"""
    )

    if no_refresh:
        logger.info("Token manager not started due to '--no-refresh' flag.")
    elif no_daemon:
        logger.info("Starting token manager in foreground...")
        start_token_manager(refresh_period, config)
    else:
        logger.info("Starting token manager daemon...")
        daemonize_token_manager(refresh_period, config)


@auth.command()
@click.option(
    "--config-file",
    type=ResolvedPath(exists=True, dir_okay=False, resolve_path=True),
    default=IQMClientCliCommand.default_config_path,
)
@click.option("--keep-tokens", is_flag=True, default=False, help="Don't delete tokens file, but kill token manager.")
@click.option("-f", "--force", is_flag=True, default=False, help="Don't ask for confirmation.")
def logout(config_file: str, keep_tokens: str, force: bool) -> None:
    """Either logout completely, or just stop token manager while keeping tokens file."""
    config = _validate_config_file(config_file)
    auth_server_url, realm, client_id = str(config.auth_server_url), config.realm, config.client_id
    tokens_file = config.tokens_file

    if not tokens_file.is_file():
        click.echo("Not logged in.")
        return

    try:
        tokens = _validate_tokens_file(str(tokens_file))
    except click.FileError:
        click.echo("Found invalid tokens.json, cannot perform any logout steps.")
        return

    pid = check_token_manager(str(tokens_file))
    refresh_token = tokens.refresh_token

    # 1. Keep tokens, kill daemon
    if keep_tokens and pid:
        if force or click.confirm("Keep tokens file and kill token manager. OK?", default=None):
            _safe_process_terminate(pid, "Token manager killed.")
            return

    # 2. Keep tokens, do nothing
    if keep_tokens and not pid:
        click.echo("Token manager is not running, and you chose to keep tokens. Nothing to do, exiting.")
        return

    # 3. Delete tokens, perform logout, kill daemon
    if not keep_tokens and pid:
        if force or click.confirm("Logout from server, delete tokens and kill token manager. OK?", default=None):
            try:
                logout_request(auth_server_url, realm, client_id, refresh_token)
            except (Timeout, ConnectionError, ClientAuthenticationError) as error:
                logger.warning(
                    "Failed to revoke tokens due to error when connecting to authentication server: %s", error
                )

            _safe_process_terminate(pid)
            os.remove(tokens_file)
            logger.info("Tokens file deleted. Logged out.")
            return

    # 4. Delete tokens, perform logout
    if not keep_tokens and not pid:
        if force or click.confirm("Logout from server and delete tokens. OK?", default=None):
            try:
                logout_request(auth_server_url, realm, client_id, refresh_token)
            except (Timeout, ConnectionError, ClientAuthenticationError) as error:
                logger.warning(
                    "Failed to revoke tokens due to error when connecting to authentication server: %s", error
                )

            os.remove(tokens_file)
            logger.info("Tokens file deleted. Logged out.")
            return

    logger.info("Logout aborted.")


def save_tokens_file(path: str, tokens: dict[str, str], auth_server_url: str) -> None:
    """Saves tokens as JSON file at given path.

    Args:
        path (str): path to the file to write
        tokens (dict[str, str]): authorization access and refresh tokens
        auth_server_url (str): base url of the authorization server
    Raises:
        OSError: if writing to file fails

    """
    path_to_dir = Path(path).parent
    tokens_data = {
        "timestamp": datetime.now().isoformat(),
        "access_token": tokens["access_token"],
        "refresh_token": tokens["refresh_token"],
        "auth_server_url": auth_server_url,
    }

    try:
        path_to_dir.mkdir(parents=True, exist_ok=True)
        with open(Path(path), "w", encoding="UTF-8") as file:
            file.write(json.dumps(tokens_data))
    except OSError as error:
        raise click.ClickException(f"Error writing tokens file, {error}") from error


def _safe_process_terminate(pid: int, msg: str = "") -> None:
    """Try to terminate a process given a process ID (PID).
    Suppress the exception in case the process is not existing
    """
    with contextlib.suppress(NoSuchProcess):
        Process(pid).terminate()
        if msg:
            logger.info(msg)


if __name__ == "__main__":
    iqmclient(sys.argv[1:])



# Copyright 2022-2025 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Qiskit adapter for IQM's quantum computers."""

from iqm.qiskit_iqm.fake_backends import IQMErrorProfile, IQMFakeAdonis, IQMFakeAphrodite, IQMFakeApollo, IQMFakeDeneb
from iqm.qiskit_iqm.fake_backends.iqm_fake_backend import IQMFakeBackend
from iqm.qiskit_iqm.iqm_circuit import IQMCircuit
from iqm.qiskit_iqm.iqm_job import IQMJob
from iqm.qiskit_iqm.iqm_move_layout import generate_initial_layout
from iqm.qiskit_iqm.iqm_naive_move_pass import IQMNaiveResonatorMoving, transpile_to_IQM
from iqm.qiskit_iqm.iqm_provider import IQMBackend, IQMProvider, __version__
from iqm.qiskit_iqm.iqm_transpilation import IQMOptimizeSingleQubitGates, optimize_single_qubit_gates
from iqm.qiskit_iqm.move_gate import MoveGate
from iqm.qiskit_iqm.transpiler_plugins import *  # noqa: F403
from qiskit import __version__ as qiskit_version



# Copyright 2020–2025 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""IQM's Aphrodite quantum architecture."""

from __future__ import annotations

from .iqm_device import IQMDevice, IQMDeviceMetadata


class Aphrodite(IQMDevice):
    r"""IQM's 54-qubit transmon device.

    The qubits are connected thus::

            QB54  QB51  QB46  QB39
            /  \  /  \  /  \  /  \
         QB53  QB50  QB45  QB38  QB31
         /  \  /  \  /  \  /  \  /
      QB52  QB49  QB44  QB37  QB30
         \  /  \  /  \  /  \  /  \
         QB48  QB43  QB36  QB29  QB22
         /  \  /  \  /  \  /  \  /  \
      QB47  QB42  QB35  QB28  QB21  QB14
         \  /  \  /  \  /  \  /  \  /
         QB41  QB34  QB27  QB20  QB13
         /  \  /  \  /  \  /  \  /  \
      QB40  QB33  QB26  QB19  QB12  QB7
         \  /  \  /  \  /  \  /  \  /
         QB32  QB25  QB18  QB11  QB6
            \  /  \  /  \  /  \  /  \
            QB24  QB17  QB10  QB5   QB2
            /  \  /  \  /  \  /  \  /
         QB23  QB16  QB9   QB4   QB1
            \  /  \  /  \  /
            QB15  QB8   QB3

    where the lines denote which qubit pairs can be subject to two-qubit gates.

    Aphrodite has native PhasedXPowGate, XPowGate, and YPowGate gates. The two-qubit gate CZ is
    native, as well. The qubits can be measured simultaneously or separately any number of times.
    """

    def __init__(self):
        qubit_count = 54
        connectivity = (
            {1, 2},
            {1, 5},
            {2, 6},
            {3, 4},
            {3, 9},
            {4, 5},
            {4, 10},
            {5, 6},
            {5, 11},
            {6, 7},
            {6, 12},
            {7, 13},
            {8, 9},
            {8, 16},
            {9, 10},
            {9, 17},
            {10, 11},
            {10, 18},
            {11, 12},
            {11, 19},
            {12, 13},
            {12, 20},
            {13, 14},
            {13, 21},
            {14, 22},
            {15, 16},
            {15, 23},
            {16, 17},
            {16, 24},
            {17, 18},
            {17, 25},
            {18, 19},
            {18, 26},
            {19, 20},
            {19, 27},
            {20, 21},
            {20, 28},
            {21, 22},
            {21, 29},
            {22, 30},
            {23, 24},
            {24, 25},
            {24, 32},
            {25, 26},
            {25, 33},
            {26, 27},
            {26, 34},
            {27, 28},
            {27, 35},
            {28, 29},
            {28, 36},
            {29, 30},
            {29, 37},
            {30, 31},
            {30, 38},
            {31, 39},
            {32, 33},
            {32, 40},
            {33, 34},
            {33, 41},
            {34, 35},
            {34, 42},
            {35, 36},
            {35, 43},
            {36, 37},
            {36, 44},
            {37, 38},
            {37, 45},
            {38, 39},
            {38, 46},
            {40, 41},
            {41, 42},
            {41, 47},
            {42, 43},
            {42, 48},
            {43, 44},
            {43, 49},
            {44, 45},
            {44, 50},
            {45, 46},
            {45, 51},
            {47, 48},
            {48, 49},
            {48, 52},
            {49, 50},
            {49, 53},
            {50, 51},
            {50, 54},
            {52, 53},
            {53, 54},
        )
        super().__init__(IQMDeviceMetadata.from_qubit_indices(qubit_count, connectivity))



# Copyright 2021-2023 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Authorization and session management for IQM Client CLI."""

from base64 import b64decode
from enum import Enum
import json
import time

from pydantic import BaseModel, Field
import requests

REFRESH_MARGIN_SECONDS = 15
AUTH_REQUESTS_TIMEOUT = 20


class ClientAuthenticationError(RuntimeError):
    """Something went wrong with user authentication."""


class ClientAccountSetupError(RuntimeError):
    """User's account has not been fully set up yet."""


class GrantType(str, Enum):
    """Type of token request."""

    PASSWORD = "password"
    REFRESH = "refresh_token"


class AuthRequest(BaseModel):
    """Request sent to authentication server for access token and refresh token, or for terminating the session.
    * Token request with grant type ``'password'`` starts a new session in the authentication server.
      It uses fields ``client_id``, ``grant_type``, ``username`` and ``password``.
    * Token request with grant type ``'refresh_token'`` is used for maintaining an existing session.
      It uses field ``client_id``, ``grant_type``, ``refresh_token``.
    * Logout request uses only fields ``client_id`` and ``refresh_token``.
    """

    client_id: str = Field(...)
    "name of the client for all request types"
    grant_type: GrantType | None = Field(None)
    "type of token request, in ``{'password', 'refresh_token'}``"
    username: str | None = Field(None)
    "username for grant type ``'password'``"
    password: str | None = Field(None)
    "password for grant type ``'password'``"
    refresh_token: str | None = Field(None)
    "refresh token for grant type ``'refresh_token'`` and logout request"


def slash_join(a: str, b: str) -> str:
    """Join two URL segments together, ensuring a single slash between them."""
    return a.rstrip("/") + "/" + b.lstrip("/")


def login_request(url: str, realm: str, client_id: str, username: str, password: str) -> dict[str, str]:
    """Sends login request to the authentication server.

    Raises:
        ClientAuthenticationError: obtaining the tokens failed

    Returns:
        Tokens dictionary

    """
    data = AuthRequest(client_id=client_id, grant_type=GrantType.PASSWORD, username=username, password=password)

    request_url = slash_join(url, f"realms/{realm}/protocol/openid-connect/token")
    result = requests.post(request_url, data=data.model_dump(exclude_none=True), timeout=AUTH_REQUESTS_TIMEOUT)
    if result.status_code == 404:
        raise ClientAuthenticationError(f"token endpoint is not available at {url}")
    if result.status_code == 400 and result.json().get("error_description", "") == "Account is not fully set up":
        raise ClientAccountSetupError("Account is not fully set up")
    if result.status_code != 200:
        raise ClientAuthenticationError("invalid username and/or password")
    tokens = result.json()
    tokens = {key: tokens.get(key, "") for key in ["access_token", "refresh_token"]}
    return tokens


def refresh_request(url: str, realm: str, client_id: str, refresh_token: str) -> dict[str, str] | None:
    """Sends refresh request to the authentication server.

    Raises:
        Timeout: no response from auth server within the timeout period
        ConnectionError: connecting the auth server failed on all retries
        ClientAuthenticationError: updating the tokens failed

    Returns:
        Tokens dictionary, or None if refresh_token is expired.

    """
    if not token_is_valid(refresh_token):
        raise ClientAuthenticationError("Refresh token has expired")

    # Update tokens using existing refresh_token
    data = AuthRequest(client_id=client_id, grant_type=GrantType.REFRESH, refresh_token=refresh_token)

    request_url = slash_join(url, f"realms/{realm}/protocol/openid-connect/token")
    result = requests.post(request_url, data=data.model_dump(exclude_none=True), timeout=AUTH_REQUESTS_TIMEOUT)
    if result.status_code != 200:
        raise ClientAuthenticationError(f"Failed to update tokens, {result.text}")
    tokens = result.json()
    if not tokens or "access_token" not in tokens or "refresh_token" not in tokens:
        raise ClientAuthenticationError("Failed to get new tokens")
    tokens = {key: tokens.get(key, "") for key in ["access_token", "refresh_token"]}
    return tokens


def logout_request(url: str, realm: str, client_id: str, refresh_token: str) -> bool:
    """Sends logout request to the authentication server.

    Raises:
        ClientAuthenticationError: updating the tokens failed

    Returns:
        True if logout was successful

    """
    data = AuthRequest(client_id=client_id, refresh_token=refresh_token)
    request_url = slash_join(url, f"realms/{realm}/protocol/openid-connect/logout")
    result = requests.post(request_url, data=data.model_dump(exclude_none=True), timeout=AUTH_REQUESTS_TIMEOUT)

    if result.status_code != 204:
        raise ClientAuthenticationError(f"Failed to logout, {result.text}")
    return True


def time_left_seconds(token: str) -> int:
    """Check how much time is left until the token expires.

    Returns:
        Time left on token in seconds.

    """
    _, body, _ = token.split(".", 2)
    # Add padding to adjust body length to a multiple of 4 chars as required by base64 decoding
    body += "=" * (-len(body) % 4)
    exp_time = int(json.loads(b64decode(body)).get("exp", "0"))
    return max(0, exp_time - int(time.time()))


def token_is_valid(refresh_token: str) -> bool:
    """Check if token is not about to expire.

    Returns:
        True if token is still valid, False otherwise.

    """
    return time_left_seconds(refresh_token) > REFRESH_MARGIN_SECONDS
